DEBUG = false
editor_backup = 3
time_scale = 1
prefs = {}

--all sprites are scaled 2x
--a single brick is 32 pixels wide and 16 pixels high
--a powerup is the same size as a brick
--a regular ball has a radius of 7 pixels
--the game board is 13 bricks wide and 32 bricks high
window = {w = 800, h = 600}
window.lwallx = 192 --(800 - 32*13) / 2
window.rwallx = window.w - window.lwallx
window.ceiling = 88
window.ceil = window.ceiling
window.wallw = window.lwallx
window.boardw = window.rwallx - window.lwallx
window.boardh = window.h - window.ceiling

--this affects paddle movement
--may be expanded so that the keyboard can navigate buttons
control_mode = "mouse" --mouse, keyboard, smart_keyboard
paddle_speed = 500 --paddle speed for keyboard only (mouse has infinite speed)
global_scale = 1
global_cursor = nil

function setScaledScissor(x, y, w, h)
	love.graphics.setScissor(x*global_scale, y*global_scale, w*global_scale, h*global_scale)
end

require("lovedebug")
class = require("middleclass")
vector = require("hump.vector") --I only used this for hardoncollider.sat
-- HC = require("hardoncollider")
shapes = require("hardoncollider.shapes")

recentTrace = "empty"
function saveTraceback()
	recentTrace = debug.traceback()
end

function printBox(box)
	print(box[1], box[2], box[3], box[4])
end

function shapes.newRectangleShape(x, y, w, h)
	return shapes.newPolygonShape(x,y, x+w,y, x+w,y+h, x,y+h)
end

function saveTest()
	print("Creating file text.txt...")
	local file = love.filesystem.newFile("test.txt")
	file:open("w")
	file:write("this is a test")
	file:close()

	local path = love.filesystem.getSaveDirectory()

	local check = love.filesystem.exists("test.txt")
	if check then 
		print("Sucessfully written to computer!")
		print("Please check "..path.." to verify existence of test.txt")
	else
		print("ERROR: File not found! I have failed you...")
	end
end

function generatePlaylist(n, start, finish)
	local file = love.filesystem.newFile("playlists/_Zone"..n)
	file:open("w")
	for i = start, finish do
		file:write("_Round"..n..i.."\n")
	end
	file:close()
end

function loadPreferences()
	local chunk = love.filesystem.load("preferences.lua")
	if chunk then
		prefs = chunk()
	end
	if not prefs.volume then prefs.volume = 20 end
	love.audio.setVolume(prefs.volume/100)
end

function savePreferences()
	local file = love.filesystem.newFile("preferences.lua")
	file:open("w")
	file:write("local data = {\n")
	for k, v in pairs(prefs) do
		file:write("\t"..k.." = "..v..",\n")
	end
	file:write("}\nreturn data")
end

shapeTypes = {"polygon", "compound", "circle", "point"}

 --not really called by any of my scripts
gameTypes = {"paddle", "ball", "brick", "projectile", "powerup", "particle", "environment"}

--this specifies the order in which the objects are updated (however it shouldn't really matter)
--paddle is updated the last
listTypes = {"balls", "bricks", "projectiles", "powerups", "callbacks", "particles", "environments", "menacers", "droppers"}

--adds an object to an object queue in order to not interfere with any iteration
--make sure its a valid list in the game

function love.load(arg)
	love.window.setMode(window.w * global_scale, window.h * global_scale,{borderless = false, vsync = false})
	love.filesystem.setIdentity("OtakuBall")
	love.window.setTitle("Otaku-Ball")
	math.randomseed(os.time())
	require("my_util")
	require("media")
	require("callback")
	require("sprite")
	require("ball")
	require("projectile")
	require("paddle")
	require("bricks/_init")
	require("powerup")
	require("particle")
	require("environment")
	require("monitor")
	require("menacer")
	require("game")
	require("gui")
	require("documentation")

	loadPreferences()

	freeze_canvas = love.graphics.newCanvas(window.w, window.h)
	shadow_canvas = love.graphics.newCanvas(window.w, window.h)

	--after this is called, the EditorState class will have two class variables: brickData and brickDataLookup
	EditorState:initBrickData()

	--controls
	mouse = {}
	keys = {}
	love.keyboard.setKeyRepeat(true)

	game:initialize()
	game:push(MainMenuState:new())
end

function love.quit()
	if editorstate then
		editorstate:close() --allows the editorstate to save backups
	end
	savePreferences()
end

--all events are done before love.update()

function love.textinput(text)
	keys.lastText = text
end

function love.keypressed(key, code, isrepeat)
	if isrepeat then
		keys[key] = 2
	else
		keys[key] = 1
	end

	--print(key.." "..code.." "..(isrepeat and "repeat" or "no repeat"))
end

function love.wheelmoved(x, y)
	if y > 0 then mouse.scrollup = true end
	if y < 0 then mouse.scrolldown = true end
end

--before the game updates, both the mouse buttons and keyboard buttons have been accounted for

function love.update(dt)
	global_cursor = nil
	--mouse events
	--m1 == 1 means mouse one is clicked
	--m1 ~= nil means mouse one is down
	local buttons = {m1 = 1, 
					 m2 = 2, 
					 m3 = 3}
	for k, v in pairs(buttons) do
		if love.mouse.isDown(v) then
			if not mouse[k] then mouse[k] = 1 else mouse[k] = 2 end
		else
			mouse[k] = false
		end
	end

	-- if dt < 1/60 then love.timer.sleep(1/60 - dt) end --fps cap for debugging reasons

	mouse.x, mouse.y = love.mouse.getPosition()
	mouse.x = mouse.x / global_scale
	mouse.y = mouse.y / global_scale

	dt = math.min(dt, 1/60) --this makes sure that the game will run at least 60 fps
	game:update(dt)

	for k, v in pairs(keys) do
		keys[k] = nil
	end
	keys.lastText = nil
	mouse.scrollup, mouse.scrolldown = nil, nil

	love.mouse.setCursor(global_cursor)
end

function love.draw()
	love.graphics.scale(global_scale)
	game:draw()
	love.graphics.setFont(font["default"])
	love.graphics.setColor(0, 0, 0, 255)
	love.graphics.setFont(font["Pokemon16"])
	love.graphics.print("FPS: "..love.timer.getFPS(), 5, window.h - 20)
	love.graphics.setColor(255, 255, 255, 255)
	-- love.graphics.setFont(font["Arcade20"])
	-- love.graphics.print("The quick brown fox jumps over the lazy dog", window.lwallx, window.h/2)
	-- love.graphics.setFont(font["Pokemon20"])
	-- love.graphics.print("The quick brown fox jumps over the lazy dog", window.lwallx, window.h/2 + 100)
end