
menacerSpawner = {timer = 2, flag = {}}

function menacerSpawner:initialize(flag)
	self.flag = flag
	if self.flag["redgreen"] then
		self.flag["red"] = true
		self.flag["green"] = true
		self.flag["redgreen"] = nil
	end
	self.timer = 2
	self.state = "waiting" --waiting, deploying
	self.gates = playstate.gates.top
	self.deployment = {}
end

function menacerSpawner:update(dt)
	self.timer = self.timer - dt
	if self.timer <= 0 then
		self.timer = math.random(2, 8)

		local names = {"cyan", "bronze", "silver", "pewter", "red", "green"}
		local menacerCount = {}
		local greenBrickCount = 0

		for i, v in ipairs(names) do
			menacerCount[v] = 0
		end
		for _, menacer in pairs(game.menacers) do
			local t = menacer.menacerType
			menacerCount[t] = menacerCount[t] + 1
		end
		for _, br in pairs(game.bricks) do
			if br.brickType == "GreenBrick" then
				greenBrickCount = greenBrickCount + 1
			end
		end

		local choices = {}
		for _, v in ipairs(names) do
			if self.flag[v] and menacerCount[v] < 3 then
				if v == "red" then
					if greenBrickCount > 0 then
						table.insert(choices, v)
					end
				else
					table.insert(choices, v)
				end
			end
		end

		if #choices > 0 then
			spawnDropper(choices[math.random(#choices)])
		end
	end
	util.remove_if(self.deployment, function(dropper)
		if dropper.gate.state == "opened" then
			game:emplace("droppers", dropper)
			return true
		end
		return false
	end)
end

function spawnDropper(menacerType)
	menacerSpawner:spawnDropper(menacerType)
end

function menacerSpawner:spawnDropper(menacerType)
	local candidates = {}
	for _, gate in ipairs(self.gates) do
		if not gate.dropper and gate.state == "closed" then
			table.insert(candidates, gate)
		end
	end
	if #candidates == 0 then
		return 0
	end
	local gate = candidates[math.random(#candidates)]
	gate.state = "opening"
	gate.target = 23*2

	local vx, vy = util.rotateVec(100, 0, math.random(10, 30))
	if math.random(1, 2) == 1 then vx = -vx end
	local x = gate.middle + 2
	local y = window.ceiling - 16 - 57/2
	local dropper = Dropper:new(x, y, vx, vy, menacerType)
	dropper.gate = gate
	gate.dropper = dropper
	table.insert(self.deployment, dropper)
end

Menacer = class("Menacer", Sprite)

lookup = {ball = {}, dropper = {}}
local names = {"red", "green", "cyan", "bronze", "silver", "pewter"}
local ref = {{0, 0}, {0, 1}, {0, 2}, {1, 0}, {1, 1}, {1, 2}}
for i, v in ipairs(names) do
	lookup.ball[v] = rects.ball2[(i-1)%3+1][(i <= 3) and 7 or 8]
	local n = ref[i]
	lookup.dropper[v] = 
	{
		before = make_rect(n[2]*(23+1),        n[1]*24, 23, 24),
		after  = make_rect(n[2]*(23+1), 2*24 + n[1]*24, 23, 24)
	}
end

function Menacer:initialize(x, y, vx, vy, menacerType)
	Sprite.initialize(self, "ball_spritesheet_new", lookup.ball[menacerType], 14, 14, x, y, vx, vy)
	self:setShape(shapes.newCircleShape(0, 0, 10))
	self.menacerType = menacerType

	self.stuckBounceTimer = 0
	self.storedAngle = 0

	self.gameType = "menacer"
end

function Menacer:update(dt)
	if self.disable then return end

	local r = self:getR()
	local x, y = self:getPos()

	self.stuckBounceTimer = self.stuckBounceTimer + dt

	if x - r < window.lwallx  then self:handleCollision( 1,  0) end
	if x + r > window.rwallx  then self:handleCollision(-1,  0) end
	if y - r < window.ceiling then self:handleCollision( 0,  1) end
	if y - r > window.h 	  then self.dead = true end

	Sprite.update(self, dt)
end

-- function Menacer:draw()
-- 	local ratio = 21.875 / 20
-- 	local w = self.w * ratio
-- 	local h = self.h * ratio
-- 	love.graphics.setColor(self.color.r, self.color.g, self.color.b, self.color.a)
-- 	draw(self.imgstr, self.rect, self.x, self.y, self.angle, w, h, nil, nil, nil, nil, self.drawFloor)
-- end

function Menacer:onBrickHit(brick, norm)
	local b = nil
	if brick.brickType == "NormalBrick" then
		b = "Normal"
	elseif brick.brickType == "MetalBrick" then
		b = brick.metalType.."Metal"
	elseif brick.brickType == "GreenBrick" then
		b = "GreenMenacer"
	end
	local m = self.menacerType
	if b and (m == "bronze" or m == "silver" or m == "green" or m == "red") then
		local check = false
		if m == "bronze" and b ~= "BronzeMetal" then
			brick:kill()
			local br = MetalBrick:new(brick.x, brick.y, 20)
			br.underlay = {imgstr = brick.imgstr, rect = brick.rect}
			br:playAnimation("BronzeCoating")
			br:inheritMovement(brick)
			game:emplace("bricks", br)
		elseif m == "silver" and b ~= "SilverMetal" then
			brick:kill()
			local br = MetalBrick:new(brick.x, brick.y, 30)
			br.underlay = {imgstr = brick.imgstr, rect = brick.rect}
			br:playAnimation("SilverCoating")
			br:inheritMovement(brick)
			game:emplace("bricks", br)
		elseif m == "green" and b ~= "GreenMenacer" then
			brick:kill()
			local br = GreenBrick:new(brick.x, brick.y)
			br.underlay = {imgstr = brick.imgstr, rect = brick.rect}
			br:playAnimation("GreenCoating")
			br:inheritMovement(brick)
			game:emplace("bricks", br)
		elseif m == "red" and b == "GreenMenacer" then
			brick:kill()
		end
	end
	self:handleCollision(norm.x, norm.y)
end

function Menacer:onDropperHit(dropper, norm)
	self:handleCollision(norm.x, norm.y)
end

function Menacer:handleCollision(xn, yn)
	--normalize
	local dist = util.dist(xn, yn)
	xn = xn / dist
	yn = yn / dist
	--check to see if collision is valid
	if not Ball.validCollision(self, xn, yn) then return false end --stealing a method from Ball
	--vector reflection
	local dot = self.vx*xn + self.vy*yn
	self.vx = self.vx - (2 * dot * xn)
	self.vy = self.vy - (2 * dot * yn)
	--another check that prevents the ball from bouncing too vertically or horizontally for a long duration
	local angle = math.abs(math.atan2(self.vy, self.vx)*180.0/math.pi)
    local vert = math.floor((angle + 45) / 90) % 2 == 1
    angle = math.min(angle, 180.0 - angle)
    angle = math.min(angle, 90.0 - angle)
    if angle > 10 or not util.deltaEqual(angle, self.storedAngle) then
    	self.stuckBounceTimer = 0
    end
    self.storedAngle = angle
    if self.stuckBounceTimer > 5 then
    	self.stuckBounceTimer = 0
    	local speed = self:getSpeed()
    	local k = vert and "vx" or "vy"
    	local d = (self[k] >= 0 and 1 or -1) * 0.3 * speed
    	self[k] = self[k] + d
    	self:scaleVelToSpeed(speed)
    end
	return true
end

function Menacer:validCollision(xn, yn)
	if xn == 0 and yn == 0 then return false end
	local theta = util.angleBetween(xn, yn, self.vx, self.vy)
	return theta > math.pi / 2
end

function Menacer:getR()
	return self.shape._radius
end

Dropper = class("Dropper", Sprite)

function Dropper:initialize(x, y, vx, vy, menacerType)
	Sprite.initialize(self, "dropper2", lookup.dropper[menacerType].before, 23*2, 24*2, x, y, vx, vy)
	self:setShape(shapes.newRectangleShape(0, 0, self.w, self.h))
	self.menacerType = menacerType
	self.hasDropped = false
	self.drawFloor = true

	self.gameType = "dropper"
end

function Dropper:destructor()
	if self.gate then
		self.gate.state = "closing"
		self.gate.dropper = nil
	end
end

function Dropper:update(dt)
	if self.disable then return end

	local x, y = self:getPos()
	local w, h = self:getDim()
	w, h = w/2, h/2
	if x - w < window.lwallx  then self.vx = math.abs(self.vx) end
	if x + w > window.rwallx  then self.vx = -math.abs(self.vx) end
	if y - h < window.ceiling then self.vy = math.abs(self.vy) end --what's the point?
	if y - h > window.h       then self:kill() end

	local p = game.paddle
	if not self.hasDropped and y - h > window.ceiling and x > p.x - p.w/2 and x < p.x + p.w/2 then
		local check = true
		for _, br in pairs(game.bricks) do
			if util.circleRectOverlap(self.x, self.y + 14, 7, br.x, br.y, br.w, br.h) then
				check = false
				break
			end
		end
		if check then
			self.hasDropped = true
			self.rect = lookup.dropper[self.menacerType].after
			local vx, vy = util.rotateVec(0, 350, math.random(-70, 70))
			local m = Menacer:new(self.x, self.y + 14, vx, vy, self.menacerType)
			game:emplace("menacers", m)
		end
	end

	local storedvx, storedvy
	local check = y - h < window.ceiling
	if check then
		storedvx, storedvy = self.vx, self.vy
		self.vx = 0
		self.vy = math.sqrt(math.pow(storedvx, 2) + math.pow(storedvy, 2))
	elseif self.gate then
		self.gate.state = "closing"
		self.gate.dropper = nil
		self.gate = nil
	end
	Sprite.update(self, dt)
	if check then
		self.vx, self.vy = storedvx, storedvy
	end
end



function Dropper:checkBallHit(ball)
	return Brick.checkBallHit(self, ball)
end

function Dropper:onBallHit(ball, norm)
	self:kill()
	ball:onDropperHit(self, norm)
end

function Dropper:checkMenacerHit(menacer)
	return Brick.checkMenacerHit(self, menacer)
end

function Dropper:onMenacerHit(menacer, norm)
	self:kill()
	menacer:onDropperHit(self, norm)
end

function Dropper:checkProjectileHit(proj)
	return Brick.checkProjectileHit(self, proj)
end

function Dropper:onProjectileHit(proj, norm)
	self:kill()
	proj:onDropperHit(self, norm)
end

function Dropper:draw()
	love.graphics.setScissor(0, window.ceiling - 16, window.w, window.h)
	Sprite.draw(self)
	love.graphics.setScissor()
end


